from pydantic import BaseSettings

from sundew.graph import Graph
from sundew.types import Cache


class Configuration(BaseSettings):
    modules: set[str] = set()
    test_graph: Graph = Graph()
    cache: Cache = Cache()


config = Configuration()



import inspect
from collections.abc import Callable
from contextlib import _GeneratorContextManager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional


def format_kwargs(kwargs: dict[str, Any]) -> str:
    formatted_kwargs = []
    for kwarg, kwval in kwargs.items():
        if isinstance(kwval, set):
            formatted_kwargs.append(f"{repr(kwarg)}: {kwval}")
        elif inspect.isfunction(kwval):
            formatted_kwargs.append(f"{repr(kwarg)}: {kwval.__name__}")
        else:
            formatted_kwargs.append(f"{repr(kwarg)}: {repr(kwval)}")
    return f"{{{', '.join(formatted_kwargs)}}}"


@dataclass(eq=True)
class FunctionName:
    simple: str
    qualified: str


@dataclass(eq=True)
class FunctionTest:
    function: Callable
    location: str = ""
    cache: bool = False
    kwargs: dict[str, Any] = field(default_factory=dict)
    patches: dict[str, Any] = field(default_factory=dict)
    returns: Optional[Any] = None
    setup: set[Callable[[], _GeneratorContextManager[Any]]] = field(default_factory=set)
    side_effects: list[Callable[[Any], bool]] = field(default_factory=list)

    @property
    def name(self) -> FunctionName:
        simple_name = self.function.__name__
        qualified_name = self.function.__qualname__
        if simple_name == qualified_name:
            module = inspect.getmodule(self.function)
            qualified_name_array = [module.__name__] if module else []
            qualified_name_array.append(simple_name)
            qualified_name = ".".join(qualified_name_array)

        return FunctionName(simple=simple_name, qualified=qualified_name)

    def formatted_returns(self) -> Optional[Any]:
        if isinstance(self.returns, str):
            return repr(self.returns).replace('"', r"\"")
        if isinstance(self.returns, Path):
            return repr(self.returns)
        return self.returns

    def __str__(self) -> str:
        test_string = ""
        test_string += f"test({self.name.simple})(\n"
        if self.cache:
            test_string += f"\tcache={self.cache},\n"
        if self.patches:
            test_string += f"\tpatches={self.patches},\n"
        if self.setup:
            test_string += f"\tsetup={self.setup},\n"
        if self.kwargs:
            test_string += f"\tkwargs={format_kwargs(self.kwargs)},\n"
        if self.setup:
            test_string += f"\tside_effects={self.side_effects},\n"
        if self.returns:
            test_string += f"\treturns={self.formatted_returns()},\n"
        test_string += ")"

        # Python formatters prefer double quotes, let's try to adhere
        return test_string.replace("'", '"')

    def __repr__(self) -> str:
        return "FunctionTest(function={}, cache={}, kwargs={}, ".format(
            self.cache,
            self.function.__name__,
            format_kwargs(self.kwargs),
        ) + "patches={}, returns={}, setup={}, side_effects={})".format(
            self.patches,
            self.formatted_returns(),
            self.setup,
            self.side_effects,
        )

    def __hash__(self) -> int:
        return hash(repr(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FunctionTest):
            return NotImplemented
        return repr(self) == repr(other)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, FunctionTest):
            return NotImplemented
        return self.function.__name__ < other.function.__name__


@dataclass
class Function:
    declaration: Callable
    tests: set[FunctionTest] = field(default_factory=set)
    usage: set[str] = field(default_factory=set)
    deps: set[str] = field(default_factory=set)

    @property
    def name(self) -> FunctionName:
        simple_name = self.declaration.__name__
        qualified_name = self.declaration.__qualname__
        if simple_name == qualified_name:
            module = inspect.getmodule(self.declaration)
            qualified_name_array = [module.__name__] if module else []
            qualified_name_array.append(simple_name)
            qualified_name = ".".join(qualified_name_array)

        return FunctionName(simple=simple_name, qualified=qualified_name)


@dataclass()
class Cache:
    cache: dict[str, dict[str, Any]] = field(default_factory=dict)

    def contains(self, function: Callable, kwargs: dict[str, Any]) -> bool:
        function_name = function.__name__
        try:
            self.cache[function_name][format_kwargs(kwargs)]
        except KeyError:
            return False
        else:
            return True

    def get_cached_value(
        self, function: Callable, kwargs: dict[str, Any]
    ) -> Any:  # noqa: ANN401
        function_name = function.__name__
        return self.cache[function_name][format_kwargs(kwargs)]

    def update(
        self, function: Callable, kwargs: dict[str, Any], returns: Any  # noqa: ANN401
    ) -> None:
        function_name = function.__name__
        if function_name not in self.cache:
            self.cache[function_name] = {}
        self.cache[function_name][format_kwargs(kwargs)] = returns



import glob
import importlib
import importlib.util
import os
import sys
from importlib.machinery import ModuleSpec
from pathlib import Path

import typer

from sundew import test
from sundew.config import config

app = typer.Typer()


@app.command()
def run(  # noqa: C901
    module: Path,
    function: str = typer.Option(  # noqa: B008
        "",
        "--function",
        "-f",
        help="Run all tests for this function.",
    ),
    auto_test_writer: bool = typer.Option(  # noqa: B008, FBT001
        False,  # noqa: FBT003
        "--auto-test-writer",
        "-a",
        help="Enable the automatic test writer.",
    ),
) -> None:
    config.modules = {"module.name"}

    class ModuleFinder:
        @classmethod
        def find_spec(
            cls, name: str, path, target=None  # noqa: all
        ) -> ModuleSpec | None:
            module_path = module
            # Back up until we get to the root of the project directory
            while "tests" in str(module_path):
                module_path = module_path.parent

            # If we're testing a module
            module_init = module_path / name / "__init__.py"
            if module_init.exists():
                spec = importlib.util.spec_from_file_location(name, module_init)
                # If there is an __init__.py
                if spec:
                    return spec

            # If we're testing a script, look around for it
            if possible_import := glob.glob(
                f"{module_path}/**/{name}*py", recursive=True
            ):
                spec = importlib.util.spec_from_file_location(
                    name, Path(possible_import[0])
                )
                if spec:
                    return spec

            return None

    # Import user's code from tests
    sys.meta_path.append(ModuleFinder)

    if module.is_dir():
        for dirpath, _, fnames in os.walk(module):
            for f in fnames:
                if f.endswith(".py"):
                    if spec := importlib.util.spec_from_file_location(
                        "module.name",
                        Path(dirpath) / f,
                    ):
                        imported_module = importlib.util.module_from_spec(spec)
                        sys.modules["module.name"] = imported_module
                        if spec.loader:
                            spec.loader.exec_module(imported_module)
                        else:
                            ...
                    else:
                        ...
    elif spec := importlib.util.spec_from_file_location("module.name", module):
        imported_module = importlib.util.module_from_spec(spec)
        sys.modules["module.name"] = imported_module
        if spec.loader:
            spec.loader.exec_module(imported_module)
        else:
            ...
    else:
        ...

    test.run(function_name=function, enable_auto_test_writer=auto_test_writer)



from sundew.types import Function, FunctionTest


class Graph:
    def __init__(self) -> None:
        self.functions: dict[str, Function] = {}

    def extend(self, other_graph: object) -> None:
        if isinstance(other_graph, Graph):
            self.functions.update(other_graph.functions)

    def add_test(self, test: FunctionTest) -> None:
        if test.name.simple not in self.functions:
            self.functions[test.name.simple] = Function(declaration=test.function)
        self.functions[test.name.simple].tests.add(test)

    def add_connection(self, node1: Function, node2: Function) -> None:
        # Assume connection is directional from A -> B
        # Add connection between node1 and node2
        if node1.name.simple not in self.functions:
            self.functions[node1.name.simple] = node1
        self.functions[node1.name.simple].deps.add(node2.name.simple)

        # Add connection between node2 and node1
        if node2.name.simple not in self.functions:
            self.functions[node2.name.simple] = node2
        self.functions[node2.name.simple].usage.add(node1.name.simple)

    def all_functions(self) -> set[str]:
        all_deps = set()
        for function_under_test in self.functions.values():
            all_deps.add(function_under_test.name.simple)
            all_deps.update(function_under_test.deps)

        return all_deps



import inspect
from collections.abc import Callable
from contextlib import ExitStack
from pathlib import Path
from typing import Any
from unittest.mock import patch

import black

from sundew.config import config
from sundew.types import FunctionTest


class DependentFunctionSpy:
    def __init__(self, func: Callable) -> None:
        self.func = func
        self.calls: set[FunctionTest] = set()
        self.__name__ = func.__name__
        self.__qualname__ = func.__qualname__

    def __call__(
        self, *args: tuple[Any, ...], **kwargs: dict[str, Any]
    ) -> Any:  # noqa: ANN401
        generated_kwargs = inspect.getcallargs(self.func, *args, **kwargs)

        # Check cache
        if config.cache.contains(self.func, generated_kwargs):
            return config.cache.get_cached_value(self.func, generated_kwargs)

        answer = self.func(*args, **kwargs)
        function_test = FunctionTest(
            function=self.func,
            kwargs=generated_kwargs,
            returns=answer,
            cache=True,
        )
        self.calls.add(function_test)
        return answer

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DependentFunctionSpy):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return (
            self.func.__code__.co_code == other.func.__code__.co_code
            and self.calls == other.calls
        )


def mock_function_dependencies(
    fn: Callable, stack: ExitStack
) -> dict[str, DependentFunctionSpy]:
    mocks: dict[str, DependentFunctionSpy] = {}
    for dep_func_simple_name in config.test_graph.functions[fn.__name__].deps:
        dep_func = config.test_graph.functions[dep_func_simple_name]
        # Recursively mock sub-dependent functions
        if (
            config.test_graph.functions[dep_func_simple_name].deps
            and dep_func_simple_name != "mock_function_dependencies"
        ):
            mocks.update(mock_function_dependencies(dep_func.declaration, stack))
        # Spy on all dependent functions
        mocks[dep_func_simple_name] = stack.enter_context(
            patch(
                dep_func.name.qualified,
                DependentFunctionSpy(dep_func.declaration),
            )
        )
    return mocks


def check_kwargs_for_imports(test_fn: FunctionTest) -> list[list[str]]:
    imports = []
    for kwval in test_fn.kwargs.values():
        if inspect.isfunction(kwval) or inspect.isclass(kwval):
            module = inspect.getmodule(kwval)
            qualified_name_array = [module.__name__] if module else []
            qualified_name_array.append(kwval.__name__)
            imports.append(qualified_name_array)
        elif isinstance(kwval, (list, tuple, set)):
            for el in kwval:
                if inspect.isfunction(el):
                    module = inspect.getmodule(el)
                    qualified_name_array = [module.__name__] if module else []
                    qualified_name_array.append(el.__name__)
                    imports.append(qualified_name_array)
                elif hasattr(el, "__slots__"):
                    module = inspect.getmodule(type(el))
                    qualified_name_array = [module.__name__] if module else []
                    qualified_name_array.append(type(el).__name__)
                    imports.append(qualified_name_array)
    return imports


def check_returns_for_imports(test_fn: FunctionTest) -> list[list[str]]:
    imports = []
    if inspect.isfunction(test_fn.returns):
        module = inspect.getmodule(test_fn.returns)
        qualified_name_array = [module.__name__] if module else []
        qualified_name_array.append(test_fn.returns.__name__)
        imports.append(qualified_name_array)
    elif isinstance(test_fn.returns, Path):
        imports.append(["pathlib", "PosixPath"])

    return imports


def generate_naive_function_import(
    mock_name: str, mock_test_functions: set[FunctionTest]
) -> list[tuple[str, str]]:
    imports = []
    # Get main function import
    imports.append(config.test_graph.functions[mock_name].name.qualified.split("."))
    # Check for any additional functions in kwargs
    for test_fn in mock_test_functions:
        imports.extend(check_kwargs_for_imports(test_fn))
        imports.extend(check_returns_for_imports(test_fn))

    return [
        (
            ".".join(function_under_test[:-1]),
            function_under_test[-1],
        )
        for function_under_test in imports
    ]


def build_test_strings(fn_tests: set[FunctionTest]) -> str:
    all_function_tests = ""
    # Sort tests alphabeticaly by function name
    for test in sorted(fn_tests):
        all_function_tests += test.__str__() + "\n"

    return all_function_tests


def build_import_string(generated_test_file_imports: set[tuple[str, str]]) -> str:
    generated_test_file_import_string = ""
    import_paren_limit = 2

    # Group naive test imports
    grouped_generated_test_file_imports: dict[str, list[str]] = {
        "sundew.test": ["test"]
    }
    for generated_from, generated_import in generated_test_file_imports:
        grouped_generated_test_file_imports.setdefault(generated_from, []).append(
            generated_import
        )

    # Build import string from groups
    for grouped_from, grouped_import in sorted(
        grouped_generated_test_file_imports.items()
    ):
        generated_test_file_import_string += "from {} import {}{}{}\n".format(
            grouped_from,
            "(\n\t" if len(grouped_import) > import_paren_limit else "",
            ",\n\t".join(sorted(grouped_import))
            if len(grouped_import) > import_paren_limit
            else ", ".join(sorted(grouped_import)),
            ",\n)" if len(grouped_import) > import_paren_limit else "",
        )
    generated_test_file_import_string += "\n"
    return generated_test_file_import_string


def write_tests_to_file(
    file_path: Path,
    generated_test_file_import_string: str,
    generated_test_file: str,
) -> None:
    with (file_path).open("w") as test_file:
        test_file.write(
            generated_test_file_import_string.expandtabs(4)
            + generated_test_file.expandtabs(4)
        )
    # Auto format file
    black.format_file_in_place(
        file_path,
        fast=True,
        mode=black.FileMode(),
        write_back=black.WriteBack.YES,
    )


def build_file_path(fn: Callable) -> Path:
    if module_path := inspect.getmodule(fn).__file__:  # type: ignore[union-attr]
        file_path = Path(module_path).resolve()
    else:
        file_path = Path(".").resolve()

    for _ in config.test_graph.functions[fn.__name__].name.qualified.split(".")[:-1]:
        file_path = file_path.parent
    file_path = file_path / "tests"
    for directory in config.test_graph.functions[fn.__name__].name.qualified.split(".")[
        1:-1
    ]:
        # Create directory
        file_path /= directory

    file_path /= f"auto_test_{fn.__name__}.py"

    return file_path


def generate_function_dependency_test_file(
    fn: Callable, mocks: dict[str, DependentFunctionSpy]
) -> None:
    generated_test_file_imports = set()
    mock_calls: set[FunctionTest] = set()
    # Build imports
    for mock_name, mock_fn in mocks.items():
        # Only add imports if we generated tests for the mocked function
        if mock_fn.calls:
            for generated_import in generate_naive_function_import(
                mock_name, mock_fn.calls
            ):
                generated_test_file_imports.add(generated_import)
            # Only add tests we don't have written already
            mock_calls.update(
                mock_fn.calls - config.test_graph.functions[mock_name].tests
            )

    generated_test_file_import_string = build_import_string(generated_test_file_imports)

    # Build file path
    file_path = build_file_path(fn)

    # Build test file all at once
    generated_test_file = build_test_strings(mock_calls)

    if generated_test_file:
        file_path.parents[0].mkdir(parents=True, exist_ok=True)
        # Add __init__.py if it doesn't exist
        (file_path.parents[0] / "__init__.py").touch()
        write_tests_to_file(
            file_path, generated_test_file_import_string, generated_test_file
        )



import ast
import inspect
from collections.abc import Callable
from typing import Any


def generate_inverse_operation_str(  # noqa: C901, PLR0911
    operation: ast.cmpop,
) -> str:
    if isinstance(operation, ast.Eq):
        return "!="
    if isinstance(operation, ast.NotEq):
        return "=="

    if isinstance(operation, ast.Lt):
        return ">="
    if isinstance(operation, ast.GtE):
        return "<"

    if isinstance(operation, ast.LtE):
        return ">"
    if isinstance(operation, ast.Gt):
        return "<="

    if isinstance(operation, ast.Is):
        return "is not"
    if isinstance(operation, ast.IsNot):
        return "is"

    if isinstance(operation, ast.In):
        return "is not in"
    if isinstance(operation, ast.NotIn):
        return "is in"
    return "unknown operator"


# Generate assert statement from a Compare node
def generate_compare_assert(node: ast.Compare) -> ast.Assert:
    return ast.Assert(
        test=node,
        msg=ast.JoinedStr(
            values=[
                ast.Constant(
                    value=f"Side-effect '{ast.unparse(node)}' failed because: "
                ),
                ast.Constant(value=f"{ast.unparse(node.left)} == "),
                ast.FormattedValue(value=node.left, conversion=-1),
                ast.Constant(value=", and "),
                ast.FormattedValue(value=node.left, conversion=-1),
                ast.Constant(value=f" {generate_inverse_operation_str(node.ops[0])} "),
                ast.FormattedValue(value=node.comparators[0], conversion=-1),
            ]
        ),
    )


# Generate a general assert statement, for expressions we don't explicit handle yet
def generate_general_assert(node: ast.expr) -> ast.Assert:
    return ast.Assert(
        test=node,
        msg=ast.JoinedStr(
            values=[
                ast.Constant(value=ast.unparse(node)),
                ast.Constant(value=" is False."),
            ],
        ),
    )


class ConvertSideEffect(ast.NodeTransformer):
    def visit_Lambda(self, node: ast.Lambda) -> ast.Assert:  # noqa: N802
        if isinstance(node.body, ast.Compare):
            return generate_compare_assert(node.body)
        return generate_general_assert(node.body)


class SideEffectGetter(ast.NodeTransformer):
    def __init__(self) -> None:
        super().__init__()
        self.lambda_sources: set[str] = set()

    def visit_Lambda(self, node: ast.Lambda) -> None:  # noqa: N802
        self.lambda_sources.add(ast.unparse(node).strip())

    def get(self, code_string: str) -> set[str]:
        tree = ast.parse(code_string)
        self.visit(tree)
        return self.lambda_sources


def get_source(funcs: list[Callable[[Any], bool]]) -> set[str]:
    side_effects = SideEffectGetter()
    for func in funcs:
        code_string = inspect.getsource(func).strip()
        side_effects.get(code_string)

    return side_effects.lambda_sources



import ast
import asyncio
import importlib
import inspect
import os
import sys
from collections.abc import Callable, Coroutine
from contextlib import ExitStack, _GeneratorContextManager
from functools import wraps
from typing import Any, Optional, Union
from unittest.mock import patch

from pydantic import BaseConfig, BaseModel, create_model
from rich import print
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeElapsedColumn,
)

from sundew.config import config
from sundew.side_effects import ConvertSideEffect, get_source
from sundew.test_writer import (
    generate_function_dependency_test_file,
    mock_function_dependencies,
)
from sundew.types import Function, FunctionTest


def update_test_graph(fn: Callable) -> None:
    func_name = fn.__name__
    # Don't check this wrapper function
    if func_name != "sundew_test_wrapper":
        # Get package_name, module_name, and module
        if module := inspect.getmodule(fn):
            module_name = module.__name__
            imported_module = importlib.import_module(module_name)
            config.modules.add(imported_module.__name__)

            # Loop through sub functions called
            for sub_func_name in fn.__code__.co_names:
                # We only care about functions in the modules being tested
                if hasattr(imported_module, sub_func_name):
                    # Verify that the funcion is native to the modules being tested
                    sub_function = getattr(imported_module, sub_func_name)
                    parent_module = getattr(
                        sub_function,
                        "__module__",
                        None,
                    )
                    # Only connect functions within the module
                    if parent_module in config.modules and inspect.isfunction(
                        sub_function
                    ):
                        config.test_graph.add_connection(
                            Function(declaration=fn),
                            Function(declaration=sub_function),
                        )
        else:
            print("Not sure how we get here yet.")


def get_test_function(fn: Callable) -> Callable:
    @wraps(fn)
    def sundew_test_wrapper(
        *args: tuple[Any, ...],
        **kwargs: dict[str, Any],
    ) -> Callable:
        return fn(*args, **kwargs)

    @wraps(fn)
    async def async_sundew_test_wrapper(
        *args: tuple[Any, ...],
        **kwargs: dict[str, Any],
    ) -> Coroutine:
        return await fn(*args, **kwargs)

    if inspect.iscoroutinefunction(fn):
        async_sundew_test_wrapper.__signature__ = inspect.signature(  # type: ignore[attr-defined]
            fn
        )
        return async_sundew_test_wrapper
    sundew_test_wrapper.__signature__ = inspect.signature(fn)  # type: ignore[attr-defined]
    return sundew_test_wrapper


def test(fn: Callable) -> Callable:
    def add_test(
        cache: bool = False,  # noqa: FBT
        setup: Optional[set[Callable[[], _GeneratorContextManager[Any]]]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        returns: Union[Any, Exception, None] = None,
        patches: Optional[dict[str, Any]] = None,
        side_effects: Optional[list[Callable]] = None,
    ) -> Callable:
        update_test_graph(fn)

        config.test_graph.add_test(
            FunctionTest(
                function=get_test_function(fn),
                kwargs=kwargs or {},
                cache=cache,
                location=f"{os.path.relpath(inspect.stack()[1][1])}:{inspect.stack()[1][2]}",
                patches=patches or {},
                returns=returns,
                setup=setup or set(),
                side_effects=side_effects or [],
            ),
        )

        return add_test

    return add_test


def apply_patches(test: FunctionTest, stack: ExitStack) -> None:
    if test.patches:
        for target, new in test.patches.items():
            stack.enter_context(patch(target, new))


def apply_fixtures(test: FunctionTest, stack: ExitStack) -> dict[str, Any]:
    fixture_yields = {}
    if test.setup:
        for fixture in test.setup:
            fixture_yield: Any = stack.enter_context(fixture())
            fixture_yields[fixture.__name__] = fixture_yield

    return fixture_yields


def copy_function_inputs(test: FunctionTest) -> dict[str, Any]:
    isolated_inputs: dict[str, Any] = {}
    signature = inspect.signature(test.function)
    isolated_inputs = {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

    for arg_name, input_value in test.kwargs.items():
        if callable(input_value):
            if inspect.isgeneratorfunction(input_value):  # type: ignore[arg-type]
                with input_value() as input_val:
                    result = input_val
            else:
                result = input_value
        else:
            result = input_value
        isolated_inputs[arg_name] = result
    return isolated_inputs


def run_function(
    test: FunctionTest, isolated_input: dict[str, Any]
) -> Any:  # noqa: ANN401
    if test.cache and config.cache.contains(test.function, test.kwargs):
        return config.cache.get_cached_value(test.function, test.kwargs)

    if inspect.iscoroutinefunction(test.function):
        return asyncio.run(test.function(**isolated_input))
    return test.function(**isolated_input)


def check_test_exception(test: FunctionTest, e: Exception) -> None:
    # If we get an exception, check to see if it was expected
    if test.returns:
        assert isinstance(e, test.returns)
    else:
        raise e


def check_test_output(test: FunctionTest, actual_return: Any) -> None:  # noqa: ANN401
    if isinstance(actual_return, ast.Module) and isinstance(test.returns, ast.Module):
        assert (
            ast.unparse(actual_return).strip() == ast.unparse(test.returns).strip()
        ), (
            f"Input {test.kwargs} returned AST for "
            + f"{ast.unparse(actual_return).strip()} "
            + "which does not match the expected AST for "
            + f"{ast.unparse(test.returns).strip()}"
        )
    #     assert actual_return.__code__.co_code == test.returns.__code__.co_code, (
    else:
        assert actual_return == test.returns, (
            f"Input {test.kwargs} returned {actual_return} "
            + "which does not match the expected return of "
            + f"{test.returns}"
        )


def build_side_effect_vars(test_function: Callable) -> type[BaseModel]:
    # Parse function signature annotations
    funtion_arguments: dict[str, tuple[Any, Any]] = {}
    for name, parameter in inspect.signature(test_function).parameters.items():
        funtion_arguments[name] = (
            # parameter or Any,
            parameter.annotation
            if parameter.annotation is not inspect.Signature.empty
            else Any,
            ...,
        )

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    return create_model(
        "SideEffectVars",
        **funtion_arguments,
        patches=(dict[str, Any], {}),
        __config__=Config,
    )  # type: ignore[call-overload]


def check_test_side_effects(
    test: FunctionTest,
    actual_return: Any,  # noqa: ANN401
    isolated_input: dict[str, Any],
) -> None:
    if test.side_effects:
        side_effect_sources = get_source(test.side_effects)
        for side_effect_source in side_effect_sources:
            side_effect_ast = ConvertSideEffect().visit(ast.parse(side_effect_source))
            side_effect_code = ast.unparse(side_effect_ast)

            side_effect_arg_model = build_side_effect_vars(test.function)
            _ = side_effect_arg_model.construct(
                patches=test.patches, returns=actual_return, **isolated_input
            )
            exec(side_effect_code)  # noqa: S102


def run_test(
    test: FunctionTest,
    stack: ExitStack,
    tests_ran: TaskID,
    progress: Progress,
    enable_auto_test_writer: bool,  # noqa: FBT001
) -> None:
    try:
        # programmatically apply any patches defined
        apply_patches(test, stack)
        # programmatically apply any fixtures setup
        apply_fixtures(test, stack)
        try:
            # Isolate input arguments
            isolated_input = copy_function_inputs(test)

            # Mock dependent functions to be able to write regression tests
            mocks = mock_function_dependencies(test.function, stack)

            # Run the test function once for evaluation
            actual_return = run_function(test, isolated_input)

            if enable_auto_test_writer:
                # Write tests with mock data
                generate_function_dependency_test_file(test.function, mocks)

        except Exception as e:  # noqa: BLE001
            # If we get an exception, check if it's expected
            check_test_exception(test, e)
        else:
            # If we didn't get an exception then check the output normally
            if test.returns:
                check_test_output(test, actual_return)
        finally:
            # Check if we have defined side effects
            check_test_side_effects(test, actual_return, isolated_input)

    except AssertionError as e:
        # Stop the progress bar
        progress.stop()
        # Log details of the failure
        print(f"\n[bold red1]FAILURE[/] {test.location} - {str(e)}")
        sys.exit(1)
    except Exception as e:  # noqa: BLE001
        # Stop the progress bar
        progress.stop()
        # Log details of the error
        progress.console.print_exception(show_locals=True)
        print(f"\n[bold orange1]ERROR[/] {test.location} - {str(e)}")
        sys.exit(1)
    else:
        if test.cache:
            config.cache.update(test.function, test.kwargs, actual_return)

        progress.advance(tests_ran)


def select_functions_to_test(function_name: str) -> list[str]:
    # Select function tests if provided
    if function_name:
        selected_functions = [function_name]
    else:  # Else get all tests
        selected_functions = [
            key
            for key in config.test_graph.functions
            if len(config.test_graph.functions[key].tests) > 0
        ]

    return selected_functions


def sort_tests(selected_functions: list[str]) -> list[FunctionTest]:
    # Kahn's algo to topologically sort test_graph
    all_nodes_added = False
    num_visited_nodes = 0
    sorted_tests: list[FunctionTest] = []
    functions_sorted = []

    while not all_nodes_added:
        for func_name in selected_functions:
            func = config.test_graph.functions[func_name]
            if len(func.deps) - num_visited_nodes == 0:
                sorted_tests.extend(func.tests)
                functions_sorted.append(func_name)
        if len(functions_sorted) == len(selected_functions):
            all_nodes_added = True
        else:
            num_visited_nodes += 1

    return sorted_tests


def detect_missing_tests(selected_functions: list[str]) -> list[str]:
    missing_tests: list[str] = []
    for function in config.test_graph.all_functions():
        if function not in selected_functions:
            function_name = config.test_graph.functions[function].name
            missing_tests.append(function_name.qualified or function_name.simple)

    return missing_tests


def run(function_name: str, enable_auto_test_writer: bool) -> None:  # noqa: FBT001
    with Progress(
        TextColumn("{task.description}"),
        BarColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
    ) as progress:
        total_num_tests = sum(
            len(func.tests) for func in config.test_graph.functions.values()
        )

        selected_functions = select_functions_to_test(function_name)
        sorted_tests = sort_tests(selected_functions)
        missing_tests = detect_missing_tests(selected_functions)

        if missing_tests:
            progress.console.print(
                f"WARNING: Missing tests detected for {missing_tests}",
            )

        progress.console.print(
            f"Selected {len(sorted_tests)}/{total_num_tests} tests",
        )

        tests_ran = progress.add_task("Running tests...", total=len(sorted_tests))

        # Run all selected tests
        for test in sorted_tests:
            with ExitStack() as stack:
                run_test(test, stack, tests_ran, progress, enable_auto_test_writer)
